import json
from datetime import datetime, timedelta
import random
from typing import Dict, List, Tuple, Any
from fastapi import FastAPI, HTTPException, Request

from schemas import TripRequest


def load_graph_from_request(request: TripRequest) -> Tuple[
    Dict[Tuple[str, str], List[Any]], Dict[Tuple[str, str], List[Any]], str, str, List[Dict], List[
        Dict], datetime, datetime]:
    graphTo = {}
    flightsDataTo = []
    graphBack = {}
    flightsDataBack = []

    # Загрузить данные о полетах из запроса
    flightsTo = request.flightsTo
    flightsBack = request.flightsBack
    start_point = request.fromPoint
    end_point = request.toPoint
    departure_date = datetime.fromisoformat(request.departureDate).date()  # Преобразование строки в объект даты
    departure_date_return = request.returnDate
    if departure_date_return is not None and departure_date_return != "":
        departure_date_return = datetime.fromisoformat(request.returnDate).date()  # Преобразование строки в объект даты

    # Добавить каждый полет в граф
    for flight in flightsTo:
        departure_point = flight.departurePoint
        destination_point = flight.destinationPoint
        departure_time = datetime.fromisoformat(flight.departureTime)
        destination_time = datetime.fromisoformat(flight.destinationTime)

        # Сохранение данных о полете для вывода
        flightsDataTo.append({
            "id": flight.id,
            "departurePoint": departure_point,
            "destinationPoint": destination_point,
            "departureTime": departure_time,
            "destinationTime": destination_time
        })

        # Проверка наличия ребра в графе и его создание при необходимости
        if (departure_point, destination_point) not in graphTo:
            graphTo[(departure_point, destination_point)] = []

        # Добавление ребра в граф
        graphTo[(departure_point, destination_point)].append([flight.distance, departure_time, destination_time])

    # Добавить каждый полет в граф
    for flight in flightsBack:
        departure_point = flight.departurePoint
        destination_point = flight.destinationPoint
        departure_time = datetime.fromisoformat(flight.departureTime)
        destination_time = datetime.fromisoformat(flight.destinationTime)

        # Сохранение данных о полете для вывода
        flightsDataBack.append({
            "id": flight.id,
            "departurePoint": departure_point,
            "destinationPoint": destination_point,
            "departureTime": departure_time,
            "destinationTime": destination_time
        })

        # Проверка наличия ребра в графе и его создание при необходимости
        if (departure_point, destination_point) not in graphBack:
            graphBack[(departure_point, destination_point)] = []

        # Добавление ребра в граф
        graphBack[(departure_point, destination_point)].append([flight.distance, departure_time, destination_time])

    return graphTo, graphBack, start_point, end_point, flightsDataTo, flightsDataBack, departure_date, departure_date_return


def path_length_and_time(path, graph, start_point, end_point, departure_date):
    if path[0] != start_point or path[-1] != end_point:
        return float('inf'), None, None

    length = 0
    if (path[0], path[1]) not in graph:
        return float('inf'), None, None

    start_time = graph[(path[0], path[1])][0][1]
    end_time = start_time

    if start_time.date() != departure_date:
        return float('inf'), start_time, end_time

    for i in range(len(path) - 1):
        if (path[i], path[i + 1]) not in graph:
            return float('inf'), start_time, end_time
        flight = graph[(path[i], path[i + 1])][0]
        length += flight[0]
        if i > 0:  # Пропуск первой вершины
            if end_time + timedelta(minutes=180) > flight[1]:
                return float('inf'), start_time, end_time  # Неприемлемый путь
        end_time = flight[2]
    return length, start_time, end_time


def generate_population(size, start, end, graph, departure_date, max_attempts=100):
    population = []
    nodes = list(set([k[0] for k in graph.keys()] + [k[1] for k in graph.keys()]))
    for i in range(size):
        attempts = 0
        while attempts < max_attempts:
            path = [start]
            visited = set(path)
            while path[-1] != end:
                next_nodes = [k[1] for k in graph.keys() if k[0] == path[-1]]
                if not next_nodes:
                    break
                next_node = random.choice(next_nodes)
                if next_node in visited:
                    break
                path.append(next_node)
                visited.add(next_node)
                if len(path) > len(nodes):
                    break
            if path[-1] == end:
                # Проверка даты вылета первого рейса
                first_leg = (path[0], path[1])
                if first_leg in graph and graph[first_leg][0][1].date() == departure_date:
                    population.append(path)
                break
            attempts += 1
    return population


def select_parents(population, graph, start_point, end_point, departure_date):
    sorted_population = sorted(population,
                               key=lambda p: path_length_and_time(p, graph, start_point, end_point, departure_date)[0])
    valid_parents = [p for p in sorted_population if
                     path_length_and_time(p, graph, start_point, end_point, departure_date)[0] != float('inf')]
    return valid_parents[:len(valid_parents) // 2]


def crossover(parent1, parent2):
    crossover_point = len(parent1) // 2
    child = parent1[:crossover_point]
    for gene in parent2:
        if gene not in child:
            child.append(gene)
    return child


def mutate(child):
    if len(child) <= 2:
        return child  # Если путь состоит из одной вершины, мутация не требуется
    index1 = random.randint(1, len(child) - 2)
    index2 = random.randint(1, len(child) - 2)
    child[index1], child[index2] = child[index2], child[index1]
    return child


def update_best_paths(best_paths, path, graph, start_point, end_point, departure_date, max_best_paths=10):
    length, _, _ = path_length_and_time(path, graph, start_point, end_point, departure_date)
    if path not in best_paths:
        best_paths[path] = length
    if len(best_paths) > max_best_paths:
        worst_path = max(best_paths, key=best_paths.get)
        del best_paths[worst_path]


def genetic_algorithm(start, end, graph, flights_data, type, departure_date, population_size=2000, generations=100):
    population = generate_population(population_size, start, end, graph, departure_date)
    best_paths = {}  # Словарь для хранения уникальных лучших маршрутов и их длин
    for generation in range(generations):
        parents = select_parents(population, graph, start, end, departure_date)

        # Если не удалось выбрать родителей, завершить алгоритм
        if not parents:
            break

        new_population = parents[:]
        while len(new_population) < population_size:
            parent1 = random.choice(parents)
            parent2 = random.choice(parents)
            child = crossover(parent1, parent2)
            if random.random() < 0.1:  # Вероятность мутации
                child = mutate(child)
            if child[-1] == end:
                new_population.append(child)
                update_best_paths(best_paths, tuple(child), graph, start, end, departure_date)
        population = new_population

    best_paths = sorted(best_paths.items(), key=lambda item: item[1])
    top_paths = [list(path) for path, _ in best_paths[:10]]

    result = []
    for path in top_paths:
        path_data = []
        for i in range(len(path) - 1):
            for flight in flights_data:
                if flight['departurePoint'] == path[i] and flight['destinationPoint'] == path[i + 1]:
                    path_data.append({
                        "id": flight['id'],
                        "departurePoint": flight['departurePoint'],
                        "destinationPoint": flight['destinationPoint']
                    })
                    break

        # Финальная проверка пути
        if path[0] == start and path[-1] == end and path_length_and_time(path, graph, start, end, departure_date)[
            0] != float('inf'):
            result.append({
                type: path_data,
            })

    return result